//
//  Cyclic graph:
//  - list and map of object pointers
//  - shared objects (several pointers pointing to the same object)
//  - cyclic graph (object A points to object B and vice-versa)
//  jsonserial - Eric Lecolinet
// See: https://www.telecom-paris.fr/~elc/jsonserial/
//

#include "main.hpp"
#include "../jsonserial.hpp"
#include "../jsonserial_list.hpp"
#include "../jsonserial_map.hpp"
using namespace jsonserial;

namespace cyclic_graph {

class Contact {
public:
  enum Gender {Unknown, Male, Female, Other};
  struct Phone {std::string type, number;};

  friend JSonClasses* registerClasses();  // needed if variables are not public!

  Contact() = default;                        // needed to instantiate pointees!

  Contact(std::string const& firstname, std::string const& lastname,
          unsigned int age, Gender gender)
  : firstname(firstname), lastname(lastname), age(age), gender(gender) {}

  void addPhone(std::string const& type, std::string const& number) {
    phones.push_back(new Phone{type, number});
  }

  void linkPartners(Contact* p) {
    partner = p;
    if (p) p->partner = this;
  }

  void linkChildWithParents(Contact* p1, Contact* p2) {
    parent1 = p1;
    parent2 = p2;
    if (p1) p1->children[firstname] = this;
    if (p2) p2->children[firstname] = this;
  }

  void print(std::ostream& out) const {
    out
    << "First:\t" << firstname << "\n"
    << "Last:\t"  << lastname  << "\n"
    << "Gender:\t"<< int(gender) << "\n"
    << "Age:\t" << age << "\n"
    << "Phones:\t";
    for (auto& it : phones) {
      out << "[" << it->type << ": " << it->number << "] ";
    }
    out << "\n";
    if (partner) out << "Partner:\t"<< partner->firstname << "\n";
    if (parent1) out << "Parent1:\t"<< parent1->firstname << "\n";
    if (parent2) out << "Parent2:\t"<< parent2->firstname << "\n";
    if (!children.empty()) {
      out << "Children:\t";
      for (auto& it : children) out << it.second->firstname << " ";
      out << "\n";
    }
  }
  
private:
  std::string firstname, lastname;
  unsigned int age{};
  Gender gender;
  std::list<Phone*> phones;                   // list of object pointers
  std::map<std::string, Contact*> children;   // map of pointed objects
  Contact *partner{}, *parent1{}, *parent2{}; // CAUTION: pointers MUST be initialized!
};

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

using Contacts = std::list<Contact*>;

void printList(std::string const& title, Contacts const& list) {
  std::cout << "--- " << title << " --------\n\n";
  for (auto& it : list) {it->print(std::cout); std::cout << std::endl;}
}

Contacts createList() {
  Contact* bob = new Contact("Bob", "Dûpontèle", 31, Contact::Male);
  bob->addPhone("home", "123 456-7890");
  bob->addPhone("mobile","703 221-2121");

  Contact* alice = new Contact("Alice", "Dûpontèle", 33, Contact::Female);
  alice->addPhone("home", "123 456-7890");
  alice->addPhone("office", "211 1234-1234");

  Contact* karim = new Contact("Karim", "Dûpontèle", 9, Contact::Male);
  karim->addPhone("mobile","122 122-1222");

  Contact* susan = new Contact("Susan", "Dûpontèle", 11, Contact::Gender::Female);
  susan->addPhone("mobile","133 133-1333");

  bob->linkPartners(alice);
  karim->linkChildWithParents(bob, alice);
  susan->linkChildWithParents(bob, alice);

  return Contacts{bob, alice, karim, susan};
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

JSonClasses* registerClasses() {
  auto* classes = new JSonClasses();

  classes->addClass<Contact>("Contact")
  .addMember("firstname", &Contact::firstname)
  .addMember("lastname", &Contact::lastname)
  .addMember("gender", &Contact::gender)
  .addMember("age", &Contact::age)
  .addMember("phones", &Contact::phones)
  .addMember("partner", &Contact::partner)
  .addMember("parent1", &Contact::parent1)
  .addMember("parent2", &Contact::parent2)
  .addMember("children", &Contact::children);

  classes->addClass<Contact::Phone>("Phone")
  .addMember("type", &Contact::Phone::type)
  .addMember("number", &Contact::Phone::number);

  return classes;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

static JSonSerial js(registerClasses());


void test_file(std::string const& filename) {
  Contacts list = createList();
  printList("Original List", list);

  // CAUTION: last argument must be true because the objects' graph is cyclic!
  if (!js.write(list, filename, true)) return;

  Contacts list2;
  if (!js.read(list2, filename)) return;
  printList("Copy after Reading", list2);
}


void test_stream() {
  Contacts list = createList();
  printList("Original List", list);

  std::stringstream buffer;

  // CAUTION: last argument must be true because the objects' graph is cyclic!
  if (!js.write(list, buffer, true)) return;
  std::cout << buffer.str() << std::endl;

  Contacts list2;
  if (!js.read(list2, buffer)) return;
  printList("Copy after Reading", list2);
}

}
